

Tables = {
    "General": {
        "ACTIVE DEGREES OF FREEDOM",
        "ANALYSIS OPTIONS",
        "PROGRAM CONTROL",
        "PROJECT INFORMATION",
        "REBAR SIZES",
    },

    "Options": {
        "OPTIONS - COLORS - DISPLAY",
        "OPTIONS - COLORS - OUTPUT",
        "TABLES AUTOMATICALLY SAVED AFTER ANALYSIS",
    },

    "Connectivity": {
        "CONNECTIVITY - AREA",
        "CONNECTIVITY - CABLE",
        "CONNECTIVITY - FRAME",
        "CONNECTIVITY - LINK",
        "CONNECTIVITY - SOLID",
        "CONNECTIVITY - TENDON",
    },

    "Loads": {
        "LOAD CASE DEFINITIONS",
        "LOAD PATTERN DEFINITIONS",
        "MASS SOURCE",
    },

    "Material": {
        "MATERIAL PROPERTIES 01 - GENERAL",
        "MATERIAL PROPERTIES 02 - BASIC MECHANICAL PROPERTIES",

        "MATERIAL PROPERTIES 03A - STEEL DATA",
        "MATERIAL PROPERTIES 03B - CONCRETE DATA",
        "MATERIAL PROPERTIES 03C - ALUMINUM DATA",
        "MATERIAL PROPERTIES 03D - COLD FORMED DATA",
        "MATERIAL PROPERTIES 03E - REBAR DATA",
        "MATERIAL PROPERTIES 03F - TENDON DATA",
        "MATERIAL PROPERTIES 03G - OTHER DATA",
        "MATERIAL PROPERTIES 03J - COUPLED NONLINEAR VON MISES DATA",

        "MATERIAL PROPERTIES 04 - USER STRESS-STRAIN CURVES",
        "MATERIAL PROPERTIES 06 - DAMPING PARAMETERS",
        "MATERIAL PROPERTIES 09 - ACCEPTANCE CRITERIA",
    },

    "Joint": {
        "JOINT ADDED MASS ASSIGNMENTS",
        "JOINT ADDED MASS BY VOLUME ASSIGNMENTS",
        "JOINT BRIDGE OBJECT FLAGS",
        "JOINT CONSTRAINT ASSIGNMENTS",
        "JOINT COORDINATES",
        "JOINT LOADS - FORCE",
        "JOINT LOADS - GROUND DISPLACEMENT",
        "JOINT LOCAL AXES ASSIGNMENTS 1 - TYPICAL",
        "JOINT MERGE NUMBER ASSIGNMENTS",
        "JOINT PATTERN ASSIGNMENTS",
        "JOINT PATTERN DEFINITIONS",
        "JOINT RESTRAINT ASSIGNMENTS",
        "JOINT SPRING ASSIGNMENTS 1 - UNCOUPLED",
        "JOINT VEHICLE RESPONSE COMPONENT OVERWRITES",
    },

    "Point": {
        "LINK BRIDGE OBJECT FLAGS",
        "LINK FREQUENCY DEPENDENT PROPERTIES 01 - GENERAL",
        "LINK FREQUENCY DEPENDENT PROPERTIES 02 - DETAILS",
        "LINK LOADS - GRAVITY",
        "LINK LOCAL AXES ASSIGNMENTS 1 - TYPICAL",
        "LINK LOCAL AXES ASSIGNMENTS 2 - ADVANCED",
        "LINK PROPERTY ASSIGNMENTS",
        "LINK PROPERTY DEFINITIONS - BRIDGE OBJECT FLAGS",
        "LINK PROPERTY DEFINITIONS 01 - GENERAL",
        "LINK PROPERTY DEFINITIONS 02 - LINEAR",
        "LINK PROPERTY DEFINITIONS 03 - MULTILINEAR",
        "LINK PROPERTY DEFINITIONS 04 - DAMPER",
        "LINK PROPERTY DEFINITIONS 05 - GAP",
        "LINK PROPERTY DEFINITIONS 06 - HOOK",
        "LINK PROPERTY DEFINITIONS 07 - RUBBER ISOLATOR",
        "LINK PROPERTY DEFINITIONS 08 - SLIDING ISOLATOR",
        "LINK PROPERTY DEFINITIONS 10 - PLASTIC (WEN)",
        "LINK PROPERTY DEFINITIONS 11 - MULTILINEAR PLASTIC",
        "LINK VEHICLE RESPONSE COMPONENT OVERWRITES",
    },

    "Frame": {
        "FRAME ADDED MASS ASSIGNMENTS",
        "FRAME AUTO MESH ASSIGNMENTS",
        "FRAME BRIDGE OBJECT FLAGS",
        "FRAME DESIGN PROCEDURES",
        "FRAME INSERTION POINT ASSIGNMENTS",
        "FRAME JOINT OFFSETS - BRIDGE OBJECT FLAGS",
        "FRAME LOCAL AXES ASSIGNMENTS 1 - TYPICAL",
        "FRAME P-DELTA FORCE ASSIGNMENTS",
        "FRAME PROPERTY MODIFIERS",
        "FRAME SPRING ASSIGNMENTS",

        # Hinges and releases
        "FRAME HINGE ASSIGNS 00 - HINGE DISTRIBUTION TYPE",
        "FRAME HINGE ASSIGNS 02 - USER DEFINED PROPERTIES",
        "FRAME HINGE ASSIGNS 09 - HINGE OVERWRITES",
        "FRAME RELEASE ASSIGNMENTS 1 - GENERAL",
        "FRAME RELEASE ASSIGNMENTS 2 - PARTIAL FIXITY",

        "FRAME LOAD TRANSFER OPTIONS",
        "FRAME LOADS - DISTRIBUTED",
        "FRAME LOADS - GRAVITY",
        "FRAME LOADS - POINT",
        "FRAME LOADS - TEMPERATURE",
        
        "FRAME OFFSET ALONG LENGTH ASSIGNMENTS",
        "FRAME OUTPUT STATION ASSIGNMENTS",
        "FRAME TENSION AND COMPRESSION LIMITS",
        "FRAME VEHICLE RESPONSE COMPONENT OVERWRITES",
        "FRAME SECTION ASSIGNMENTS",

        "FRAME SECTION PROPERTIES - BRIDGE OBJECT FLAGS",
        "FRAME SECTION PROPERTIES 01 - GENERAL",
        "FRAME SECTION PROPERTIES 02 - CONCRETE COLUMN",
        "FRAME SECTION PROPERTIES 05 - NONPRISMATIC",
        "FRAME SECTION PROPERTIES 06 - POLYGON DATA",
        "FRAME SECTION PROPERTIES 13 - TIME DEPENDENT",
    },

    "Shell": {
        "AREA LOCAL AXES ASSIGNMENTS 1 - TYPICAL",
        "AREA LOCAL AXES ASSIGNMENTS 2 - ADVANCED",
        "AREA ADDED MASS ASSIGNMENTS",

        "AREA AUTO MESH ASSIGNMENTS",
        "AREA BRIDGE OBJECT FLAGS",
        "AREA EDGE CONSTRAINT ASSIGNMENTS",
        "AREA LOADS - GRAVITY",
        "AREA LOADS - PORE PRESSURE",
        "AREA LOADS - ROTATE",
        "AREA LOADS - SURFACE PRESSURE",
        "AREA LOADS - TEMPERATURE",
        "AREA LOADS - UNIFORM",

        "AREA OVERWRITES - JOINT OFFSETS",
        "AREA OVERWRITES - THICKNESS",

        "AREA SECTION ASSIGNMENTS",
        "AREA SECTION PROPERTIES - BRIDGE OBJECT FLAGS",
        "AREA SECTION PROPERTIES",
        "AREA SECTION PROPERTY - TIME DEPENDENT",
        "AREA SECTION PROPERTY DESIGN PARAMETERS",
        "AREA SPRING ASSIGNMENTS",
        "AREA STIFFNESS MODIFIERS",
    },

    "Solid": {
        "SOLID AUTO MESH ASSIGNMENTS",
        "SOLID LOADS - GRAVITY",
        "SOLID LOADS - PORE PRESSURE",
        "SOLID LOADS - SURFACE PRESSURE",
        "SOLID LOADS - TEMPERATURE",
        "SOLID LOCAL AXES ASSIGNMENTS 1 - TYPICAL",
        "SOLID PROPERTY ASSIGNMENTS",
        "SOLID PROPERTY DEFINITIONS",
        "SOLID SPRING ASSIGNMENTS",
    },

    "Cable": {
        "CABLE OUTPUT STATION ASSIGNMENTS",
        "CABLE SECTION ASSIGNMENTS",
        "CABLE SECTION DEFINITIONS",
        "CABLE SHAPE DATA",
        "CABLE LOADS - DISTRIBUTED",
        "CABLE LOADS - TEMPERATURE",
    },
    "Bridge": {
        "BRIDGE ABUTMENT DEFINITIONS",
        "BRIDGE BEARING DEFINITIONS",
        "BRIDGE BENT DEFINITIONS 1 - GENERAL",
        "BRIDGE BENT DEFINITIONS 2 - COLUMN DATA",
        "BRIDGE DESIGN PREFERENCES - AASHTO LRFD 2020",
        "BRIDGE DESREQSUPER 01 - GENERAL",
        "BRIDGE DESREQSUPER 02 - STATION RANGES",
        "BRIDGE DESREQSUPER 03 - DEMAND SETS",
        "BRIDGE DESREQSUPER 04 - LLDF - AASHTOLRFD20",
        "BRIDGE DESREQSUPER 05 - PARAM - ALL CODES - CBOXTENDON",
        "BRIDGE DESREQSUPER 09 - PARAM - AASHTOLRFD20 - CBOX2STRESS",
        "BRIDGE DESREQSUPER 10 - PARAM - AASHTOLRFD20 - CBOX2SHEAR",
        "BRIDGE DESREQSUPER 11 - PARAM - AASHTOLRFD20 - CBOX2FLEXURE",
        "BRIDGE DIAPHRAGM DEFINITIONS",
        "BRIDGE ERECTION DEFINITION 01 - TRAVELER",
        "BRIDGE GROUP DEFINITION 01 - GENERAL",
        "BRIDGE GROUP DEFINITION 02 - SECTION",
        "BRIDGE GROUP DEFINITION 08 - SUPPORT STRUCTURE",
        "BRIDGE GROUP DEFINITION 09 - BEARING - SUPPORT",
        "BRIDGE GROUP DEFINITION 14 - MIXED",
        "BRIDGE LAYOUT LINE 1 - GENERAL",
        "BRIDGE LAYOUT LINE 2 - HORIZONTAL LAYOUT DATA",
        "BRIDGE LAYOUT LINE 3 - VERTICAL LAYOUT DATA",
        "BRIDGE LOAD DEFINITIONS 02 - LINE",
        "BRIDGE LOAD DEFINITIONS 03 - AREA",
        "BRIDGE OBJECT DEFINITIONS 01 - GENERAL",
        "BRIDGE OBJECT DEFINITIONS 02 - REFERENCE LINE",
        "BRIDGE OBJECT DEFINITIONS 03 - SPANS 1 - GENERAL",
        "BRIDGE OBJECT DEFINITIONS 06 - ABUTMENTS",
        "BRIDGE OBJECT DEFINITIONS 07 - BENTS",
        "BRIDGE OBJECT DEFINITIONS 09 - SUPER ELEVATION 1 - GENERAL",
        "BRIDGE OBJECT DEFINITIONS 11 - PRESTRESS 1 - GENERAL",
        "BRIDGE OBJECT DEFINITIONS 12 - PRESTRESS 2 - VERTICAL LAYOUT",
        "BRIDGE OBJECT DEFINITIONS 13 - PRESTRESS 3 - HORIZONTAL LAYOUT",
        "BRIDGE OBJECT DEFINITIONS 14 - DIAPHRAGMS",
        "BRIDGE OBJECT DEFINITIONS 15 - UPDATE DATA",
        "BRIDGE OBJECT DEFINITIONS 21 - TEMPERATURE LOADS",
        "BRIDGE OBJECT DEFINITIONS 26 - LINE LOADS",
        "BRIDGE OBJECT DEFINITIONS 27 - AREA LOADS",
        "BRIDGE OBJECT DEFINITIONS 28 - SPAN SEGMENTS",
        "BRIDGE OBJECT DEFINITIONS 31 - WIND LOADS",
        "BRIDGE OBJECT DEFINITIONS 32 - WIND LOAD - LOADPATTERN LIST",
        "BRIDGE PREFERENCES",
        "BRIDGE RATEREQSUPER 01 - GENERAL",
        "BRIDGE RATEREQSUPER 02 - STATION RANGES",
        "BRIDGE RATEREQSUPER 03 - DEMAND SETS",
        "BRIDGE RATEREQSUPER 07 - PARAM - AASHTORATE18 - CBOX2SHEAR",
        "BRIDGE RATEREQSUPER 08 - PARAM - AASHTORATE18 - CBOX2FLEXURE",
        "BRIDGE RATEREQSUPER 09 - PARAM - AASHTORATE18 - CBOX2MINREBAR",
        "BRIDGE RATEREQSUPER 20 - PARAM - AASHTORATE18 - CBOX2SERV",
        "BRIDGE RATING PREFERENCES - AASHTO RATING 2018",
        "BRIDGE RESPONSE",
        "BRIDGE SECTION CUTS 01 - GENERAL",
        "BRIDGE SECTION CUTS 02 - GROUPS",
        "BRIDGE SECTION CUTS 03 - STRESS POINTS",
        "BRIDGE SECTION CUTS 04 - GIRDER DATA - GENERAL",
        "BRIDGE SECTION CUTS 05 - GIRDER DATA - GROUPS",
        "BRIDGE SECTION DEFINITIONS - GIRDER OUTPUT LOCATION DATA",
        "BRIDGE SECTION DEFINITIONS 02 - CONCRETE BOX GIRDER",
        "BRIDGE SECTION LOAD DEFINITIONS 05 - BARRIER",
        "BRIDGE SECTION LOAD DEFINITIONS 07 - WEARING SURFACE",
        "BRIDGE SECTION LOAD DEFINITIONS 08 - TEMPERATURE CHANGE",
        "BRIDGE SEISMIC DESIGN PREFERENCES - AASHTO SEISMIC 2011",
    },
    "Other": {
        "AUTO WAVE 3 - WAVE CHARACTERISTICS - GENERAL",
        "AUTO WIND - AASHTO 2020",
        "AUTO WIND - LIVE LOAD",
        "AUTO WIND ANGLES",
        "AUTO WIND EXPOSURE FOR HORIZONTAL DIAPHRAGMS",

        "CASE - BUCKLING 1 - GENERAL",
        "CASE - BUCKLING 2 - LOAD ASSIGNMENTS",
        "CASE - DIRECT HISTORY 1 - GENERAL",
        "CASE - DIRECT HISTORY 2 - LOAD ASSIGNMENTS",
        "CASE - DIRECT HISTORY 3 - VISCOUS PROPORTIONAL DAMPING",
        "CASE - DIRECT HISTORY 4 - INTEGRATION PARAMETERS",
        "CASE - DIRECT HISTORY 5 - NONLINEAR PARAMETERS",
        "CASE - DIRECT HISTORY 6 - MODAL PROPORTIONAL DAMPING",
        "CASE - MODAL 1 - GENERAL",
        "CASE - MODAL 3 - LOAD ASSIGNMENTS - RITZ",
        "CASE - MODAL HISTORY 1 - GENERAL",
        "CASE - MODAL HISTORY 2 - LOAD ASSIGNMENTS",
        "CASE - MODAL HISTORY 4 - PROPORTIONAL DAMPING",
        "CASE - MODAL HISTORY 5 - DAMPING OVERRIDES",
        "CASE - MODAL HISTORY 6 - NONLINEAR PARAMETERS",
        "CASE - MOVING LOAD 1 - LANE ASSIGNMENTS",
        "CASE - MOVING LOAD 2 - LANES LOADED",
        "CASE - MOVING LOAD 3 - MULTILANE FACTORS",
        "CASE - MULTISTEP STATIC 1 - LOAD ASSIGNMENTS",
        "CASE - RESPONSE SPECTRUM 1 - GENERAL",
        "CASE - RESPONSE SPECTRUM 2 - LOAD ASSIGNMENTS",
        "CASE - STATIC 1 - LOAD ASSIGNMENTS",
        "CASE - STATIC 2 - NONLINEAR LOAD APPLICATION",
        "CASE - STATIC 4 - NONLINEAR PARAMETERS",
        "CASE - STATIC 5 - NONLINEAR STAGE DEFINITIONS",
        "CASE - STATIC 6 - NONLINEAR STAGE DATA",
        "CASE - STEADY STATE 1 - GENERAL",
        "CASE - STEADY STATE 2 - LOAD ASSIGNMENTS",
        "CASE - STEADY STATE 3 - ADDED FREQUENCIES GENERAL",
        "CASE - STEADY STATE 5 - ADDED SPECIFIED FREQUENCIES",
        "CASE - STEADY STATE 6 - CONSTANT DAMPING",
        "COMBINATION DEFINITIONS",

        "CONSTRAINT DEFINITIONS - BODY",
        "CONSTRAINT DEFINITIONS - BRIDGE OBJECT FLAGS",
        "CONSTRAINT DEFINITIONS - DIAPHRAGM",
        "CONSTRAINT DEFINITIONS - EQUAL",
        "CONSTRAINT DEFINITIONS - ROD",
        "COORDINATE SYSTEMS",
        "DATABASE DOCUMENTATION",
        "DATABASE FORMAT TYPES",
        "FLOATING LANE DEFINITION DATA",

        "FUNCTION - PLOT FUNCTIONS",
        "FUNCTION - POWER SPECTRAL DENSITY - USER",
        "FUNCTION - RESPONSE SPECTRUM - AASHTO 2007",
        "FUNCTION - RESPONSE SPECTRUM - USER",
        "FUNCTION - STEADY STATE - USER",
        "FUNCTION - TIME HISTORY - FROM FILE",
        "FUNCTION - TIME HISTORY - RAMP",
        "FUNCTION - TIME HISTORY - SINE",
        "FUNCTION - TIME HISTORY - USER",

        "GENERALIZED DISPLACEMENT DEFINITIONS 1 - TRANSLATIONAL",
        "GENERALIZED DISPLACEMENT DEFINITIONS 2 - ROTATIONAL",
        "GRID LINES",

        "GROUPS 1 - DEFINITIONS",
        "GROUPS 2 - ASSIGNMENTS",
        "GROUPS 4 - BRIDGE OBJECT FLAGS",

        "HINGES DEF 02 - NONINTERACTING - DEFORM CONTROL - GENERAL",
        "HINGES DEF 03 - NONINTERACTING - DEFORM CONTROL - FORCE-DEFORM",
        "HINGES DEF 04 - NONINTERACTING - DEFORM CONTROL - ACCEPTANCE",

        "LANE DEFINITION DATA",

        "NAMED SETS - DATABASE TABLES 1 - GENERAL",
        "NAMED SETS - DATABASE TABLES 2 - SELECTIONS",
        "OVERWRITES - CONCRETE DESIGN - ACI 318-08/IBC2009",
        "OVERWRITES - STEEL DESIGN - AISC360-05-IBC2006",

        "PREFERENCES - ALUMINUM DESIGN - AA 2015",
        "PREFERENCES - COLD FORMED DESIGN - AISI-16",
        "PREFERENCES - COLD FORMED DESIGN - AISI-ASD96",
        "PREFERENCES - CONCRETE DESIGN - AASHTO LRFD 2014",
        "PREFERENCES - CONCRETE DESIGN - ACI 318-08/IBC2009",
        "PREFERENCES - CONCRETE SHELL DESIGN - EUROCODE 2-2004",
        "PREFERENCES - DIMENSIONAL",
        "PREFERENCES - STEEL DESIGN - AASHTO-LRFD-2020",
        "PREFERENCES - STEEL DESIGN - AISC360-05-IBC2006",

        "SECTION CUTS 1 - GENERAL",

        "TENDON BRIDGE OBJECT FLAGS",
        "TENDON LAYOUT DATA 01 - GENERAL",
        "TENDON LAYOUT DATA 02 - SEGMENTS",
        "TENDON LOADS - TENSION FORCE OR STRESS",
        "TENDON SECTION ASSIGNMENTS",
        "TENDON SECTION DEFINITIONS - BRIDGE OBJECT FLAGS",
        "TENDON SECTION DEFINITIONS",

        "VEHICLES 2 - GENERAL VEHICLES 1 - GENERAL",
        "VEHICLES 3 - GENERAL VEHICLES 2 - LOADS",
        "VEHICLES 4 - VEHICLE CLASSES",
    }
}



#
import matplotlib.pyplot as plt
import matplotlib.tri as tri


class PlaneArtist:
    def __init__(self, model, **kwds):

        import matplotlib.pyplot as plt
        _,self.ax = plt.subplots()

        self.model = model

    def _draw_nodes(self, nodes):
        self.ax.scatter(*zip(*nodes.values()))
        for k,v in nodes.items():
            self.ax.annotate(k, v)

    def draw(self):
        pass

    def show(self):
        import matplotlib.pyplot as plt
        plt.show()


def _plot_grid(x,y, ax=None, **kwargs):
    ax = ax or plt.gca()
    segs1 = np.stack((x,y), axis=2)
    segs2 = segs1.transpose(1,0,2)
    ax.add_collection(LineCollection(segs1, **kwargs))
    ax.add_collection(LineCollection(segs2, **kwargs))

    ax.autoscale()


# converts quad elements into tri elements
def _quads_to_tris(quads):
    tris = [
        [None for j in range(3)] for i in range(2*len(quads))
    ]
    for i in range(len(quads)):
        j = 2*i
        tris[j][0] = quads[i][0]
        tris[j][1] = quads[i][1]
        tris[j][2] = quads[i][2]
        tris[j + 1][0] = quads[i][2]
        tris[j + 1][1] = quads[i][3]
        tris[j + 1][2] = quads[i][0]
    return tris


# plot edges from a finite element mesh
def _draw_plane_mesh(nodes, elements, ax):
    for element in elements:
        x = [nodes[element[i]][0] for i in range(len(element))]
        y = [nodes[element[i]][1] for i in range(len(element))]
        ax.fill(x, y, edgecolor='black', ls="-", lw=0.5, fill=False)


def render(mesh, solution, ax=None,
         # mesh options
         show_edges=True,
         # contour options
         show_scale=True
    ):
    #
    # Extract mesh information
    #
    quads = []
    elements_tris = []

    if isinstance(mesh, tuple):
        nodes, elems = mesh
        quads = quads + [
#            tuple(i-1 for i in elem) for elem in elems.values() if len(elem) == 4
             tuple(i   for i in elem) for elem in elems.values() if len(elem) == 4
        ]

    else:
        # assume mesh is a meshio object
        nodes = {i: list(coord) for i, coord in enumerate(mesh.points)}
        for blk in mesh.cells:
            if blk.type == "triangle":
                elements_tris = elements_tris + [
                    tuple(int(i) for i in elem) for elem in blk.data
                ]
            elif blk.type == "quad":
                quads = quads + [
                    tuple(int(i) for i in elem) for elem in blk.data
                ]


    elements = elements_tris + quads


    #
    # Set up canvas
    #
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.figure


    #
    # plot the finite element mesh
    #
    if show_edges:
        _draw_plane_mesh(nodes, elements, ax=ax)


    #
    # Plot solution contours
    #
    nodes_x, nodes_y = zip(*nodes.values())

    # convert all elements into triangles
    node_tag_to_index = {tag: i for i, tag in enumerate(nodes.keys())}
    elements_all_tris = [
            tuple(node_tag_to_index[tag] for tag in elem)
            for elem in elements_tris + _quads_to_tris(quads)
    ]

    # create an unstructured triangular grid instance
    triangulation = tri.Triangulation(nodes_x, nodes_y, elements_all_tris)
    contours = \
        ax.tricontourf(triangulation, solution, cmap="twilight", alpha=0.5)

    if show_scale:
        plt.colorbar(contours, ax=ax)
    ax.axis('equal')
    return ax




"""
https://github.com/pawlowiczf/Polygon-Triangulation/tree/main
"""
class Node:
    def __init__(self, color, next = None):
        self.next  = None 

        
def _create_links(colors):
    array = [Node(color) for color in colors]

    for idx in range( len(array) ):
        pointer = array[idx]
        pointer.next = array[ ( idx + 1 ) % len(array) ]
    #
        
    return array[0]



def Orientation(a, b, c):
    return (a[0] - c[0]) * (b[1] - c[1]) - (a[1] - c[1]) * (b[0] - c[0])


def Position(a, b, c):
    #
    ax, ay = a 
    bx, by = b 
    cx, cy = c 

    answer = (a[0] - c[0]) * (b[1] - c[1]) - (a[1] - c[1]) * (b[0] - c[0])

    if answer > 0:
        return 1 
    if answer < 0:
        return -1
    return float('inf')

def _point_equal(a, b):
    return a[0] == b[0] and a[1] == b[1]

def onTriangle(a, b, c, p):
    return _point_equal(a,p) or _point_equal(b,p) or _point_equal(c,p)

def inTriangle(a, b, c, p):
    #
    number = 0
    number += Position(a, b, p)
    number += Position(b, c, p)
    number += Position(c, a, p)

    if abs( number ) == 3:
        return True
    #
    return False

def _find_ear(polygon, vis):
    #
    n = len(polygon)
    if n < 3:
        return []
    
    for i in range( n  ):
        #
        a = polygon[ (i - 1) % n ][0]
        b = polygon[ i % n ][0]
        c = polygon[ (i + 1) % n ][0]

        if Orientation(a, b, c) > 0:
            ear = True 

            for p, index in polygon:
                if not onTriangle(a,b,c,p) and inTriangle(a, b, c, p):
                    ear = False 
                    break

            if ear:    
                vis.append([a, b, c])
                return ( polygon[ (i - 1) % n ][1], i % n , polygon[ (i + 1) % n ][1] )

    return []

def earcut(vertices, vis=None):
    #
    if len(vertices) == 0:
        return []

    if vis is None: 
        vis = []

    polygon = [(vertices[i], i) for i in range(len(vertices))]
    
    while len(polygon) >= 3:
        #
        index = _find_ear(polygon, vis)
        if index == []: 
            break
        polygon.pop( index[1] )

    return vis


#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
# Command line parsing
#
import os
import ast
import sys
import json
from veux import config
from veux.errors import RenderError


__version__ = "0.0.13"

NAME = "sees"

HELP = """
usage: {NAME} <sam-file>
       {NAME} --setup ...
       {NAME} [options] <sam-file>
       {NAME} [options] <sam-file> <res-file>
       {NAME} --section <py-file>#<py-object>

Generate a plot of a structural model.

Positional Arguments:
  <sam-file>                     JSON file defining the structural model.
  <res-file>                     JSON or YAML file defining a structural
                                 response.

Options:
  DISPLACEMENTS
  -s, --scale  <scale>           Set displacement scale factor.
  -d, --disp   <node>:<dof>...   Apply a unit displacement at node with tag
                                 <node> in direction <dof>.
  VIEWING
  -V, --view   {{elev|plan|sect}}  Set camera view.
      --vert   <int>             Specify index of model's vertical coordinate
      --hide   <object>          Hide <object>; see '--show'.
      --show   <object>          Show <object>; accepts any of:
                                    {{origin|frames|frames.displ|nodes|nodes.displ|extrude}}

  BACKEND
  --canvas <canvas>              trimesh, gnu, plotly, matplotlib
  --port <port>                  port to serve rendering

  MISC.
  -o, --save   <out-file>        Save plot to <out-file>.
      --conf   <conf-file>
  -c           <canvas>
      --version                  Print version and exit.
  -h, --help                     Print this message and exit.



  <dof>        {{long | tran | vert | sect | elev | plan}}
               {{  0  |   1  |   2  |   3  |   4  |   5 }}
  <object>     {{origin|frames|frames.displ|nodes|nodes.displ}}
    origin
    frames
    nodes
    legend
    extrude                      extrude cross-sections
    outline                      outline extrusion
    triads
    x,y,z

    fibers
"""


AXES = dict(zip(("long","tran","vert",
                 "sect","elev", "plan"), range(6)))

def dof_index(dof: str):
    try: return int(dof)
    except: return AXES[dof]


def sketch_show(artist_config, argn: str, arg="show", exclusive: bool=False, preserve: set=None):

    if argn == "reference" and "reference" not in artist_config["sketches"]:
        artist_config["sketches"]["reference"] = config.SketchConfig(artist_config["sketches"]["default"])
        return

    objects = argn

    if ":" in objects:
        sketches, objects = objects.split(":")
        sketches = [sketches]
    else:
        sketches = artist_config["sketches"].keys()

    if arg != "style":
        objects = objects.split(",")
    else:
        objects = [objects]

    for sketch in sketches:
        if sketch not in artist_config["sketches"]:
            artist_config["sketches"][sketch] = config.SketchConfig(artist_config["sketches"]["default"])

        for feature in objects:
            try:
                obj, feature = feature.split(".", 1)
            except:
                raise RenderError(f"Failed to parse argument {argn}; unknown object {feature}")

            if exclusive:
                for k, v in artist_config["sketches"][sketch][obj].items():
                    if preserve is not None and (sketch, obj, k) in preserve:
                        continue
                    if v is not None:
                        v["show"] = False

            if arg == "style":
                feature, val = feature.split("=")
                try:
                    feature, prop = feature.split(".")
                except ValueError:
                    raise RenderError(f"Failed to parse argument {argn}; cannot split {feature}")
                try:
                    val = ast.literal_eval(val)
                except Exception as e:
                    # If we fail to parse the value as a literal,
                    # we settle for a string
                    pass

                if  hasattr(artist_config["sketches"][sketch][obj][feature]["style"], prop):
                    setattr(artist_config["sketches"][sketch][obj][feature]["style"], prop, val)
                else:
                    artist_config["sketches"][sketch][obj][feature][prop] = val

                artist_config["sketches"][sketch][obj][feature]["show"] = True

            else:
                artist_config["sketches"][sketch][obj][feature]["show"] = (arg == "show")

            if preserve is not None:
                preserve.add((sketch, obj, feature))


def parse_args(argv)->dict:
    # 1. Defaults
    opts = config.Config()

    # 2. Local configuration file
    if os.path.exists(".render.yaml"):
        import yaml
        with open(".render.yaml", "r") as f:
            presets = yaml.load(f, Loader=yaml.Loader)

        config.apply_config(presets,opts)

    # 3. Command line

    canvas_config = opts["canvas_config"]
    artist_config = opts["artist_config"]

    args = iter(argv[1:])
    for arg in args:
        try:
            if arg == "--help" or arg == "-h":
                print(HELP.format(NAME=NAME))
                return None

            elif arg[:2] == "-o":
                filename = arg[2:] if len(arg) > 2 else next(args)
                opts["write_file"] = filename
                if "html" in filename or "json" in filename:
                    opts["canvas_config"]["type"] = "plotly"
                elif "glb" in filename or "gltf" in filename:
                    opts["canvas_config"]["type"] = "gltf"

            elif arg == "--version":
                print(__version__)
                return None

            elif arg == "--conf":
                with open(next(args), "r") as f:
                    presets = yaml.load(f, Loader=yaml.Loader)
                config.apply_config(presets,opts)

            elif arg == "--set":
                k,v = next(args).split("=")
                val = ast.literal_eval(v)
                d = opts
                keys = k.split(".")
                for key in keys[:-1]:
                    if key in d:
                        d = d[key]
                    elif f"{key}_config" in d:
                        d = d[f"{key}_config"]
                    else:
                        raise RenderError(f"Unknown config key {key}.")

                # TODO: implement type casting here
                d[keys[-1]] = val

#           elif hasattr(artist_config, "--" + arg):
#               pass

            #
            # Viewer
            #
            elif arg == "--viewer":
                opts["viewer_config"]["name"] = next(args)
            elif arg == "--port":
                opts["server_config"]["port"] = int(next(args))

            #
            # Canvas
            #
            elif arg == "--canvas":
                opts["canvas_config"]["type"] = next(args)
            elif arg == "--gnu":
                opts["canvas_config"]["type"] = "gnu"
            elif arg == "--plotly":
                opts["canvas_config"]["type"] = "plotly"

            #
            # Artist
            #
            elif arg == "--vert":
                opts["artist_config"]["vertical"] = int(next(args))

            elif arg in {"--sketch", "--reference", "--displaced"}:
                if arg == "--sketch":
                    # --sketch <sketch>.frame.outline.color=black
                    sketch, *fields = next(args).split(".")
                    if sketch not in sketches:
                        sketch = "default"
                else:
                    sketch = arg[2:]
                    fields = next(args).split(".")

            elif arg in {"--show", "--hide", "--style"}:
                # --show sk1,sk2:frame.outline
                # --size sk1,sk2:frame.outline.basis=Lagrange
                sketch_show(artist_config, next(args), arg[2:])
                continue

                objects = argn = next(args)
                if objects == "reference":
                    artist_config["sketches"]["reference"] = config.SketchConfig(artist_config["sketches"]["default"])
                    continue


                if ":" in objects:
                    sketches, objects = objects.split(":")
                else:
                    sketches = artist_config["sketches"].keys()

                if arg != "--style":
                    objects = objects.split(",")
                else:
                    objects = [objects]

                for sketch in sketches:
                    if sketch not in artist_config["sketches"]:
                        artist_config["sketches"][sketch] = config.SketchConfig(artist_config["sketches"]["default"])

                    for stroke in objects:
                        try:
                            obj, stroke = stroke.split(".", 1)
                        except:
                            raise RenderError(f"Failed to parse argument {argn}; unknown object {stroke}")

                        if arg == "--style":
                            stroke, val = stroke.split("=")
                            try:
                                stroke, prop = stroke.split(".")
                            except ValueError:
                                raise RenderError(f"Failed to parse argument {argn}; cannot split {stroke}")
                            try:
                               #val = json.loads(val)
                                val = ast.literal_eval(val)
                            except Exception as e:
                                print(e, val)
                                pass
                            if hasattr(artist_config["sketches"][sketch][obj][stroke]["style"], prop):
                                setattr(artist_config["sketches"][sketch][obj][stroke]["style"], prop, val)
                            else:
                                artist_config["sketches"][sketch][obj][stroke][prop] = val
                            artist_config["sketches"][sketch][obj][stroke]["show"] = True
#                           print(f"{sketch = }", f"{obj = }", f"{stroke = }", artist_config["sketches"][sketch][obj][stroke], sep="\n", end="\n---\n")
                        else:
                            artist_config["sketches"][sketch][obj][stroke]["show"] = (arg == "--show")



            elif arg[:2] == "-V":
                opts["view"] = arg[2:] if len(arg) > 2 else next(args)
            elif arg == "--view":
                opts["view"] = next(args)

            elif arg == "--default-section":
                opts["default_section"] = np.loadtxt(next(args))

            elif arg == "--extrude-default":
                opts["model_config"]["extrude_default"] = next(args)

            elif arg == "--extrude-outline":
                # outline used for everything
                opts["model_config"]["extrude_outline"] = next(args)


            #
            # STATE
            #
            elif arg[:2] == "-s":
                opts["scale"] = float(arg[2:]) if len(arg) > 2 else float(next(args))

            elif arg == "--scale":
                scale = next(args)
                if "=" in scale:
                    # Used like --scale <object>=<scale>
                    k,v = scale.split("=")
                    opts["objects"][k]["scale"] = float(v)
                else:
                    opts["scale"] = float(scale)

            elif arg[:2] == "-m":
                opts["mode_num"] = int(arg[2]) if len(arg) > 2 else int(next(args))

            elif arg == "--time":
                opts["state_config"]["time"] = json.loads(next(args))

            elif arg == "--recover":
                opts["state_config"]["recover"] = next(args)

            elif arg[:2] == "-d":
                node_dof = arg[2:] if len(arg) > 2 else next(args)
                for nd in node_dof.split(","):
                    node, dof = nd.split(":")
                    opts["displ"][int(node)].append(dof_index(dof))

            elif arg[:6] == "--disp":
                node_dof = next(args)
                for nd in node_dof.split(","):
                    node, dof = nd.split(":")
                    opts["displ"][int(node)].append(dof_index(dof))


            # Final check on options
            elif arg[0] == "-" and len(arg) > 1:
                raise RenderError(f"unknown option '{arg}'")

            #
            # Positional
            #
            elif not opts["sam_file"]:
                if arg == "-": arg = sys.stdin
                opts["sam_file"] = arg

            else:
                if arg == "-": arg = sys.stdin
                opts["res_file"] = arg

        except StopIteration:
            # `next(args)` was called in parse loop without successive arg
            raise RenderError(f"Argument '{arg}' expected value")

    return opts



# Claudio Perez
# Summer 2024
import textwrap
from pathlib import Path

class Viewer:
    def __init__(self, viewer=None, src=None):
        self._glbsrc = "./model.glb" if src is None else src
        self._viewer = viewer if viewer is not None else "mv"

    def get_html(self):
        if self._viewer == "three-160":
            with open(Path(__file__).parents[0]/"gltf.html", "r") as f:
                return f.read()

        elif self._viewer == "three-130":
            with open(Path(__file__).parents[0]/"index.html", "r") as f:
                return f.read()

        elif self._viewer == "mv":
            html = textwrap.dedent(f"""
          <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
            "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
          <html xmlns="http://www.w3.org/1999/xhtml" lang="en">
          <head>
            <meta charset="utf-8">
            <meta name="viewport" content="width=device-width, initial-scale=1">
            <title>veux</title>
            <script type="module" src="https://ajax.googleapis.com/ajax/libs/model-viewer/3.5.0/model-viewer.min.js">
            </script>
          </head>
          <body>
            <model-viewer alt="rendering"
                          src="{self._glbsrc}"
                          style="width: 100%; height: 500px;"
                          max-pixel-ratio="2"
                          shadow-intensity="1"
                          environment-image="/black_ground.hdr"
                          environment-image="neutral"
                          shadow-light="10000 10000 10000"
                          exposure="0.8"
                          camera-controls
                          min-camera-orbit="auto auto 0m"
                          touch-action="pan-y">
            </model-viewer>
          </body>
          </html>
        <!--
                          bounds="0 0 0 10000 10000 10000"
                          ar
                          environment-image="https://modelviewer.dev/shared-assets/environments/spruit_sunrise_1k_LDR.jpg"
                          shadow-softness="0.8"
                          ar-modes="scene-viewer; quick-look"
                          position="0 0 -50"
                          scale="0.01 0.01 0.01"
        -->
        """)
        return html




# Claudio Perez
# Summer 2024
import sys
import bottle
from veux.viewer import Viewer
from veux.server import Server


if __name__ == "__main__":

    options = {
        "viewer": None
    }
    filename = sys.argv[1]

    with open(filename, "rb") as f:
        glb = f.read()

    Server(glb=glb, **options).run()





#!/bin/env python
import sys
import math


def _list_equal(l1:list, l2:list, verbose=False)->bool:
    match = True
    for _i,_j in zip(l1,l2):
        if match != True:
            break
        else:
            i,j = _i,_j

        if isinstance(i,list):
            if isinstance(j,list):
                match = _list_equal(i,j)
            else:
                match = False

        elif isinstance(i,dict):
            if isinstance(j,dict):
                match = _dict_equal(i,j)
            else:
                match = False

        elif isinstance(i,float):
            match = match and math.isclose(i,j)

        else:
            match = match and (i==j)

    if verbose and match != True:
        print(f"{verbose}{i}\n\t{j}")

    return match


def _dict_equal(d1:bool, d2:bool,verbose=False)->bool:
    match = True

    for _k,_v in d1.items():
        if match != True: break
        else: k,v = _k,_v

        if k in {"name", "nodes", "instances", "crdTransformation"}:
            continue
        elif isinstance(v,dict):
            match = match and _dict_equal(v, d2[k])
        elif isinstance(v,list):
            match = match and _list_equal(v, d2[k])
        elif isinstance(v,float):
            match = match and math.isclose(v,d2[k])
        else:
            match = match and (v == d2[k])

    if verbose and not match:
        print(f"{verbose}{k}: {v}, {d2[k]}")

    return match

def sections(dat, mat_remap, sec_remap=None, sec_types = None):
    sec_types = [] if sec_types is None else sec_types
    sec_remap = {} if sec_remap is None else sec_remap
    secs = dat["StructuralAnalysisModel"]["properties"]["sections"]

    for el in secs:
        match = False
        if "section" in el and el["section"] in sec_remap:
            el["section"] = sec_remap[el["section"]]

        if "materials" in el:
            el["materials"] = [mat_remap[m] for m in el["materials"]]

        if "fibers" in el:
            for fib in el["fibers"]:
                fib["material"] = mat_remap[fib["material"]]

        # Test equality
        for typ in sec_types:
            if el["type"] == typ["type"] and _dict_equal(el,typ)==True:
                match = True
                break

        if not match:
            sec_types.append(el)
            el["instances"] = [el["name"]]
            sec_remap[el["name"]] = el["name"]

        else:
            sec_remap[el["name"]] = typ["name"]
            #if el in sec_types:
            #    sec_types.pop(sec_types.index(el))

    print(f"sections:\t{len(secs)}\t{len(sec_types)}", file=sys.stderr)
    return sec_remap, sec_types

def transfms(dat, crd_remap=None, crd_types = None):
    crd_types = [] if crd_types is None else crd_types
    crd_remap = {} if crd_remap is None else crd_remap
    secs = dat["StructuralAnalysisModel"]["properties"]["crdTransformations"]
    for el in secs:
        match = False
        # Test equality
        for typ in crd_types:
            if el["type"] == typ["type"] and _dict_equal(el,typ)==True:
                match = True
                break

        if not match:
            crd_types.append(el)
            el["instances"] = [el["name"]]
            crd_remap[el["name"]] = el["name"]

        else:
            crd_remap[el["name"]] = typ["name"]
            #if el in crd_types:
            #    crd_types.pop(crd_types.index(el))

    print(f"transforms:\t{len(secs)}\t{len(crd_types)}", file=sys.stderr)
    return crd_remap, crd_types

def materials(dat):
    mat_types = []
    mat_remap = {}
    mats = dat["StructuralAnalysisModel"]["properties"]["uniaxialMaterials"]
    for el in mats:
        match = False
        for typ in mat_types:
            if el["type"] == typ["type"] and _dict_equal(el,typ)==True:
                match = True
                break
        if not match:
            mat_types.append(el)
            el["instances"] = [el["name"]]
            mat_remap[el["name"]] = el["name"]
        else:
            mat_remap[el["name"]] = typ["name"]

    print(f"materials:\t{len(mats)}\t{len(mat_types)}", file=sys.stderr)
    return mat_remap, mat_types

def consolidate(model, modify = False):
    mat_remap, mat_types = materials(model)
    sec_remap, sec_types = sections(model, mat_remap)
    sec_remap, sec_types = sections(model, mat_remap, sec_remap, sec_types)
    crd_remap, crd_types = transfms(model)

    elem_types = []
    elems = model["StructuralAnalysisModel"]["geometry"]["elements"]
    for el in elems:
        match = False
        verbose  = False
        #print(f"{el['name']}")

        # change references
        if "materials" in el:
            el["materials"] = [mat_remap[m] for m in el["materials"]]

        if "sections" in el:
            el["sections"]  = [sec_remap[m] for m in el["sections"]]

        if "crdTransformation" in el:
            el["crdTransformation"] = crd_remap[el["crdTransformation"]]

        for typ in elem_types:
            if el["type"] == typ["type"] and _dict_equal(el,typ,verbose)==True:
                match = True
                break

        if not match:
            elem_types.append(el)
            el["instances"] = [el["name"]]
        else:
            typ["instances"].append(el["name"])

    model["StructuralAnalysisModel"]["properties"]["element_types"] = elem_types

    print(f"elements:\t{len(elems)}\t{len(elem_types)}",file=sys.stderr)

    return model


if __name__=="__main__":
    import json,sys
    with open(sys.argv[-1],"r") as f:
        model = json.load(f)

    clean = consolidate(model)

    print(json.dumps(clean,indent=2))




# Claudio Perez
import sys
import warnings
import numpy as np
from scipy.spatial.transform import Rotation

import shps.curve

import veux
import veux.frame
from veux.utility.earcut import earcut
from veux.model import read_model
from veux.errors import RenderError
from veux.config import MeshStyle


def draw_extrusions(model, canvas, state=None, config=None):
    #
    #     x-------o---------o---------o
    #   /       /         /
    # x--------o<--------o---------o
    # |        |       / ^
    # |        |     /   |
    # |        |   /     |
    # |        | /       |
    # x--------o-------->o---------o
    #
    ndm = 3

    coords = [] # Global mesh coordinates
    triang = [] # Triangle indices into coords
    caps   = []
    locoor = [] # Local mesh coordinates, used for textures

    if config is None:
        config = {
                "style": MeshStyle(color="gray")
        }

    I = 0
    # Track outlines with excessive edges (eg, circles) to later avoid showing
    # their edges
    no_outline = set()
    for i,el in enumerate(model["assembly"].values()):

        outline = model.cell_section(el["name"])
        if outline is None:
            continue

        outline_scale = config["scale"]

        nen  = len(el["nodes"])

        noe = len(outline) # number of outline edges
        if state is not None:
            glob_displ = state.cell_array(el["name"], state.position)
            X = shps.curve.displace(el["crd"], glob_displ, nen).T
            R = state.cell_array(el["name"], state.rotation)
        else:
            outline = outline*0.99
            outline_scale *= 0.99
            X = np.array(el["crd"])
            R = [model.frame_orientation(el["name"]).T]*nen


        try:
            caps.append(I+np.array(earcut(model.cell_section(el["name"], 0)[:,1:])))
            caps.append(I+(nen-1)*noe + np.array(earcut(model.cell_section(el["name"], 1)[:,1:])))
        except Exception as e:
            warnings.warn(f"Earcut failed with message: {e}")


        # Loop over sample points along element length to assemble
        # `coord` and `triang` arrays
        for j in range(nen):
            outline = model.cell_section(el["name"], j).copy() # TODO: Pass float between 0 and 1 instead of j
            outline[:,1:] *= outline_scale
            # Loop over section edges
            for k,edge in enumerate(outline):
                # Append rotated section coordinates to list of coordinates
                coords.append(X[j, :] + R[j]@edge)
                locoor.append(
                             [ (j+0)/nen+0.1,  0.1+(k+0)/(noe+0) ]
                )


                if j == 0:
                    # Skip the first section
                    continue

                elif k < noe-1:
                    triang.extend([
                        # Tie two triangles to this edge
                        [I+    noe*j + k,   I+    noe*j + k + 1,    I+noe*(j-1) + k],
                        [I+noe*j + k + 1,   I+noe*(j-1) + k + 1,    I+noe*(j-1) + k]
                    ])
                else:
                    # elif j < N-1:
                    triang.extend([
                        [I+    noe*j + k,    I + noe*j , I+noe*(j-1) + k],
                        [      I + noe*j, I + noe*(j-1), I+noe*(j-1) + k]
                    ])

                if len(outline) > 25:
                    no_outline.add(len(triang)-1)
                    no_outline.add(len(triang)-2)

        I += nen*noe

    triang = [list(reversed(i)) for i in triang]

    if len(triang) == 0:
        return

    cbuf, tbuf, *_ = canvas.plot_mesh(coords, triang, local_coords=locoor, style=config["style"])

    if len(caps) > 0:
        for cap in caps:
            canvas.plot_mesh(cbuf, cap, style=config["style"])


    IDX = np.array((
        (0, 2),
        (0, 1)
    ))

    triang = [list(reversed(i)) for i in triang]

    nan = np.zeros(ndm)*np.nan
    coords = np.array(coords)
    if "tran" in config["outline"]:
        tri_points = np.array([
            coords[idx]  if (j+1)%3 else nan
            for j,idx in enumerate(np.array(triang).reshape(-1))
        ])
    elif "long" in config["outline"]:
        tri_points = np.array([
            coords[i]  if j%2 else nan
            for j,idx in enumerate(np.array(triang)) for i in idx[IDX[j%2]] if j not in no_outline
        ])
    else:
        return

    canvas.plot_lines(tri_points,
                      style=config["line_style"]
    )

class so3:
    @classmethod
    def exp(cls, vect):
        return Rotation.from_rotvec(vect).as_matrix()

def _add_moment(artist, loc, axis):
    import meshio
    mesh_data = meshio.read(veux.assets/'chrystals_moment.stl')
    coords = mesh_data.points

    coords = np.einsum('ik, kj -> ij',  coords,
                       so3.exp([0, 0, -np.pi/4])@so3.exp(axis))
    coords = 1e-3*coords + loc
#   for node in coords:
#       node = so3.exp(axis)@node
    for i in mesh_data.cells:
        if i.type == "triangle":
            triangles =  i.data #mesh_data.cells['triangle']
            break

    artist.canvas.plot_mesh(coords, triangles)


def _render(sam_file, res_file=None, **opts):
    # Configuration is determined by successively layering
    # from sources with the following priorities:
    #      defaults < file configs < kwds 

    config = veux.config.Config()


    if sam_file is None:
        raise RenderError("Expected positional argument <sam-file>")

    # Read and clean model
    if not isinstance(sam_file, dict):
        model = read_model(sam_file)
    else:
        model = sam_file

    if "RendererConfiguration" in model:
        veux.apply_config(model["RendererConfiguration"], config)

    veux.apply_config(opts, config)

    artist = veux.FrameArtist(model, **config)

    draw_extrusions(artist.model, artist.canvas, config=opts)

    # -----------------------------------------------------------

    soln = veux.state.read_state(res_file, artist.model, **opts)
    if soln is not None:
        if "time" not in opts:
            soln = soln[soln.times[-1]]

        draw_extrusions(artist.model, artist.canvas, soln, opts)
        # -----------------------------------------------------------
        _add_moment(artist,
                    loc  = [1.0, 0.0, 0.0],
                    axis = [0, np.pi/2, 0])
        # -----------------------------------------------------------

    artist.draw()
    return artist


if __name__ == "__main__":
    import veux.parser
    config = veux.parser.parse_args(sys.argv)

    try:
        artist = _render(**config)

        # write plot to file if output file name provided
        if config["write_file"]:
            artist.save(config["write_file"])


        # Otherwise either create popup, or start server
        elif hasattr(artist.canvas, "popup"):
            artist.canvas.popup()

        elif hasattr(artist.canvas, "to_glb"):
            import veux.server
            server = veux.server.Server(glb=artist.canvas.to_glb(),
                                        viewer=config["viewer_config"].get("name", None))
            server.run(config["server_config"].get("port", None))

        elif hasattr(artist.canvas, "to_html"):
            import veux.server
            server = veux.server.Server(html=artist.canvas.to_html())
            server.run(config["server_config"].get("port", None))

    except (FileNotFoundError, RenderError) as e:
        # Catch expected errors to avoid printing an ugly/unnecessary stack trace.
        print(e, file=sys.stderr)
        print("         Run '{NAME} --help' for more information".format(NAME=sys.argv[0]), file=sys.stderr)
        sys.exit()




"""
https://github.com/fuhao050403/EarClippingAlgorithm
"""
import math
import sys
import numpy as np

# ========================================================== #
# Vertices processing                                        #
# Input: 1. one outer polygon vertices list                  #
#        2. all inner polygon vertices lists                 #
# Output: an simple polygon which combine all inner polygons #
#         with original outer polygon, which be used for     #
#         performing ear clipping triangulation algorithm.   #
# ========================================================== #

def vertices_processing(outer_polygon, inner_polygon):
    # Storing max-x coordinate for calculate mutually visible vertices
    index_point_M = []
    # Storing max-x values for sorting purpose
    max_x_values = []
    for i in range(len(inner_polygon)):
        # Get first column(x axis value) of inner polygon
        row_x = [row[0] for row in inner_polygon[i]]
        # Get max value of first column
        max_x_value = max(row_x)

        # Append to the max-x value/coordinate arrays
        max_x_values += [max_x_value]
        index_point_M.append(row_x.index(max_x_value))
    # Sort that array by descending order
    sort_index = np.argsort(-1 * np.array(max_x_values)).tolist()

    # Reorder inner polygon by max x-value
    # which make polygon that has greater max x-value up to the top of the list
    inner_polygon = [inner_polygon[i] for i in sort_index]
    index_point_M = [index_point_M[i] for i in sort_index]

    for i in range(len(inner_polygon)):
        # Get index of mutual visible vertex on outer polygon for current inner polygon
        index_point_P = _find_mutual_vertices(outer_polygon.copy(), inner_polygon[i][index_point_M[i]])
        
        # Insert list which contains inner polygon information about to insert into outer
        insert_list = _shift_list(inner_polygon[i], -1 * index_point_M[i])
        insert_list += [inner_polygon[i][index_point_M[i]]]
        insert_list += [outer_polygon[index_point_P]]
        
        # Insert into outer polygon
        # Combine inner polygon into outer polygon as an new outer polygon
        outer_polygon[index_point_P + 1:index_point_P + 1] = insert_list

    return outer_polygon

def _find_mutual_vertices(outer, point_M):
    len_outer = len(outer)

    # ======================================================== #
    # Find index of intersected edge and coordinate of point I #
    # ======================================================== #

    # Get right-most x value for outer polygon
    outer_rightmost_x = max([row[0] for row in outer])

    # Line segment from M to x direction (p1 = point_M)
    p2 = [outer_rightmost_x, point_M[1]]

    # index_closest record the index of first vertex for the closest edge
    index_closest = -1
    x_smallest = outer_rightmost_x + 1
    point_I = []
    for j in range(len_outer):
        p3 = outer[j]
        p4 = outer[(j + 1) % len_outer]

        # Calculate intersected point between two lines
        # and check if this point is closer to point M
        coord_intersected = _line_segments_intersect(point_M, p2, p3, p4)

        if coord_intersected != None and coord_intersected[0] < x_smallest:
            x_smallest = coord_intersected[0]
            index_closest = j
            point_I = coord_intersected

    # ============================================================= #
    # Process after found point I and index of the edge it belongs  #
    # And try to find point P and eventually find index of mutually #
    # visible vertex on the outer polygon and return that value.    #
    # ============================================================= #

    # index of mutual visible vertex on the outer polygon
    index_visible_vertex = -1

    # First. Check if the intersected point is a vertex of outer polygon
    # If so, return result index value
    index_visible_vertex = outer.index(point_I) if point_I in outer else -1
    if index_visible_vertex != -1:
        return index_visible_vertex

    # Otherwise, intersected point is an interior point of an edge
    # Find point P on that edge
    point_P = (outer[index_closest]
               if outer[index_closest][0] > outer[(index_closest + 1) % len_outer][0]
               else outer[(index_closest + 1) % len_outer])

    # ============================================================= #
    # Now, triangle <M,I,P> has been decided. So we need to check   #
    # if any reflex vertices of outer polygon is located inside it. #
    # ============================================================= #

    # Pre-set result index as index of point P on the outer polygon
    index_visible_vertex = (index_closest
                            if outer[index_closest][0] > outer[(index_closest + 1) % len_outer][0]
                            else (index_closest + 1) % len_outer)

    minimum_angle = _angle_between_vectors(point_M, point_I, point_P)
    minimum_distance = _distance_between_points(point_M, point_P)

    for k in range(len_outer):
        # If current vertex is inside <M,I,P>
        if _is_point_inside(outer[k], point_M, point_I, point_P):
            # If current vertex is reflex and is not point P
            if _is_reflex(outer[k - 1], outer[k], outer[(k + 1) % len_outer]) and outer[k] != point_P:
                # Calculate angle and distance between candidate point and point_M
                angle = _angle_between_vectors(point_M, point_I, outer[k])
                distance = _distance_between_points(point_M, outer[k])

                # If outer[k] is located exactly on line <M,P> then choose the closer one
                if angle == minimum_angle and distance < minimum_distance:
                    index_visible_vertex = k
                    minimum_distance = distance
                # If not then choose the one with minimum angle
                elif angle < minimum_angle:
                    index_visible_vertex = k
                    minimum_angle = angle

    return index_visible_vertex

# Intersection between line(p1, p2) and line(p3, p4)
def _line_segments_intersect(p1, p2, p3, p4):
    x1, y1 = p1
    x2, y2 = p2
    x3, y3 = p3
    x4, y4 = p4
    denom = (y4 - y3) * (x2 - x1) - (x4 - x3) * (y2 - y1)
    if denom == 0: # parallel
        return None
    ua = ((x4 - x3) * (y1 - y3) - (y4 - y3) * (x1 - x3)) / denom
    if ua < 0 or ua > 1: # out of range
        return None
    ub = ((x2 - x1) * (y1 - y3) - (y2 - y1) * (x1 - x3)) / denom
    if ub < 0 or ub > 1: # out of range
        return None
    x = x1 + ua * (x2 - x1)
    y = y1 + ua * (y2 - y1)
    return [x, y]

# Calculate angle between two vectors
def _angle_between_vectors(p_center, p1, p2):
    v1 = [a - b for a, b in zip(p1, p_center)]
    v2 = [a - b for a, b in zip(p2, p_center)]
    angle = np.arccos(np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2)))
    return 180 * angle/ np.pi

# Calculate distance between two points
def _distance_between_points(p1, p2):
    return (p2[0] - p1[0])**2 + (p2[1] - p1[1])**2

# Shifting list to left or right
def _shift_list(seq, n = 0):
    a = n % len(seq)
    return seq[-a:] + seq[:-a]

def earcut(polygon):
    """
    Ear Clipping triangulation algorithm
       
    Input: an polygon vertices list which gained from Vertices
           Processing part(previously mentioned)              
    Output: list of index of faces which triangulates input   
            polygon vertices.                                 

    """


    # Backup vertices array for finding index
    polygon_backup = polygon.copy()

    if len(polygon) == 3:
        return [[1, 2, 3]]

    faces = []
    while(len(polygon) > 3):
        # Found a ear
        ear_index = _find_ear(polygon)
        if(ear_index != -1):
            # Found a ear and find index of it in the backup vertices array
            try:
                prev_ind = polygon_backup.index(polygon[ear_index - 1])
                curr_ind = polygon_backup.index(polygon[ear_index])
                next_ind = polygon_backup.index(polygon[(ear_index + 1) % len(polygon)])
                faces.append([prev_ind, curr_ind, next_ind])
            except ValueError:
                print("value has not been found!")
        polygon.pop(ear_index)

    # While loop upabove ends when 3 elements left in polygon array
    # So append the last triangle manually
    try:
        prev_ind = polygon_backup.index(polygon[-1])
        curr_ind = polygon_backup.index(polygon[0])
        next_ind = polygon_backup.index(polygon[1])
        faces.append([prev_ind, curr_ind, next_ind])
    except ValueError:
        print("value has not been found!")

    return faces

# Find first possible ear of polygon
def _find_ear(polygon):
    for i in range(len(polygon)):
        prev_point = polygon[i - 1]
        curr_point = polygon[i]
        next_point = polygon[(i + 1) % len(polygon)]
        # Check if current point is a ear
        if _is_ear(prev_point, curr_point, next_point, polygon):
            # If ear, then return its index
            return i
        else:
            # If not ear, continue the loop find next possible ear
            continue
    return -1

def _is_ear(p1, p2, p3, polygon):
# Check if ths specific point is a ear
# Ear has to be satified following 3 conditions:
# 1. No other points are located inside the triangle which formed by 3 points
# 2. Center point of 3 points(p2 in this case) has a concave curve
# 3. All 3 points are not on the same lane(area of triangle must greater than 0)
    ear = _contains_no_points(p1, p2, p3, polygon) and \
          _is_convex(p1, p2, p3) and \
          _triangle_area(p1[0], p1[1], p2[0], p2[1], p3[0], p3[1]) > 0
    return ear

# Check center point of 3 points(curr in this case) has a convex curve
def _is_convex(prev, curr, next):
    return (prev[0] * (next[1] - curr[1]) + 
            curr[0] * (prev[1] - next[1]) + 
            next[0] * (curr[1] - prev[1])) > 0

# Check center point of 3 points(curr in this case) has a reflex curve
def _is_reflex(prev, curr, next):
    return (prev[0] * (next[1] - curr[1]) + 
            curr[0] * (prev[1] - next[1]) + 
            next[0] * (curr[1] - prev[1])) < 0

# Check no other points are located inside the triangle which formed by 3 points
def _contains_no_points(p1, p2, p3, polygon):
    for pn in polygon:
        if pn in (p1, p2, p3):
            continue
        elif _is_point_inside(pn, p1, p2, p3):
            return False
    return True

# Check specific point is located inside the triangle
EPSILON = math.sqrt(sys.float_info.epsilon)
def _is_point_inside(p, a, b, c):
    area  = _triangle_area(a[0], a[1], b[0], b[1], c[0], c[1])
    area1 = _triangle_area(p[0], p[1], b[0], b[1], c[0], c[1])
    area2 = _triangle_area(p[0], p[1], a[0], a[1], c[0], c[1])
    area3 = _triangle_area(p[0], p[1], a[0], a[1], b[0], b[1])
    areadiff = abs(area - sum([area1, area2, area3])) < EPSILON
    return areadiff

# Calculate area of the triangle formed by 3 points
def _triangle_area(x1, y1, x2, y2, x3, y3):
    return abs((x1 * (y2 - y3) + x2 * (y3 - y1) + x3 * (y1 - y2)) / 2.0)